%----
% Copyright (C) 1996, Gregor Meyer, FernUniversitaet Hagen, Germany.
% This file is part of the system 'Typical'.
% See the file LICENSE for conditions of distribution and use.
%----
%:- multifile (subtype)/1, (suptype)/1, (type)/1, (pred)/1, (type_import)/1, (local)/1.
















:- ensure_loaded(tc_andor).
:- ensure_loaded(typeops). 
 
:- type_import(prolog_types).
:- type_import(prolog_num).
:- type_import(prolog_meta).
:- type_import(glib).
:- type_import(csppo).
:- type_import(ti_2po).
:- type_import(ti_cl).
:- type_import(tinf_more).
:- type_import(tc_andor).
 
:- type dfg_clause = top.
 
:- local pred pragma_option(top).
 
:- pred type_infer_cl(poset(_),list(dfg_clause),list(top),list(top),list(top)). 

type_infer_cl(_,[],_,_,_).
type_infer_cl(Lat,[Cl|Cls],Sorts,Preds,Funcs) :-
	transform_func(Cl,Sorts,Preds,Funcs,Clause0),
	bind_vars(Clause0,[],Clause1),
	numbervars(Clause1,0,_),
	true,
        (pragma_option(verbose_flag) ->
         (nl,write('check/infer: '),write(Clause1),nl);
	  true),
        (true
        ;true,fail),
        %
	andor_list(Clause1,Clause2),
	or_assoc(Clause2,Clause3),
	transform_clause(Clause3,Clause4),
	true,
	reset_num_counter,
	findall(Y,aos1(Clause4,_X,Y),Poss),
	type_inference(Lat,Poss,_CO,_VT),!,
        true,
        true,
        type_infer_cl(Lat,Cls,Sorts,Preds,Funcs).

type_infer_cl(Lat,[Ill|Cls],Sorts,Preds,Funcs) :-
        % error message already written, but no source formula written
	write('of source formula '),write(Ill),nl,
        write('Press Enter to continue'),
        get0(_XXX),
        type_infer_cl(Lat,Cls,Sorts,Preds,Funcs).
 
:- pred type_inference(poset(_),list(dfg_clause),list(top),list(top)). 

type_inference(_Lat,[],[],[]).
type_inference(Lat,[Clause|ClauseList],[Co|CoList],[VT|VTList]) :-
	make_f_pairs_const(#,Clause,o,C1),
	(ti_cl(Lat,C1,CLC),
	 type_inf_more(CLC,Co,VT) ->
	 (!, 
	  (pragma_option(verbose) ->
	   (write('Formula part '), write(Clause),
	    write(' consistent with variable types '),write(VT),nl);
	   true),
	  type_inference(Lat,ClauseList,CoList,VTList));
	(nl,write('sort error caused by formula part '),
	 write(Clause),nl,
	 fail)).


% bind_vars(+Term,+VariableBindings,-NewTerm)
%
% NewTerm is the same as Term but all variables from forall and 
% exists variable lists are substituted by real Prolog variables.
% The substitution is protocolled in VariableBindings 

:- pred bind_vars(top,list(top),top).

bind_vars(Term,_VarBind,Term) :-
	var(Term),!.

bind_vars(forall(VariableList,Term),VarBind,forall(VarList2,NewTerm)) :-
	generate_vb(VariableList,VarBind,VarBind2),!,
	list_vars(VariableList,VarBind2,VarList2),
	bind_vars(Term,VarBind2,NewTerm),!.

bind_vars(exists(VariableList,Term),VarBind,exists(VarList2,NewTerm)) :-
	generate_vb(VariableList,VarBind,VarBind2),!,
	list_vars(VariableList,VarBind2,VarList2),
	bind_vars(Term,VarBind2,NewTerm),!.

bind_vars([Term],VarBind,[NewTerm]) :-
	bind_vars(Term,VarBind,NewTerm),!.

bind_vars([Term|TermList],VarBind,[NewTerm|NewTermList]) :-
	bind_vars(Term,VarBind,NewTerm),
	bind_vars(TermList,VarBind,NewTermList).

bind_vars(Term,VarBind,NewTerm) :-
	simple(Term),
	find_vb(Term,VarBind,NewTerm),!.

bind_vars(Term,_VarBind,Term) :-
	simple(Term),!.

bind_vars(Term,VarBind,NewTerm) :-
	Term =.. [Pred|Args],!,
	bind_vars(Args,VarBind,NewArgs),
	NewTerm =.. [Pred|NewArgs],!.

% list_vars(+TermList,+VariableBindings,-NewTermList)
%
% NewTermList is the same as TermList but all terms are replaced by a pair
% Term/Variable witch is extracted from VariableBindings

:- pred list_vars(list(top),list(top),list(top)).

list_vars([],_VarBind,[]).
list_vars([Term|TermList],VarBind,[Term/NewTerm|NewTermList]) :-
	find_vb(Term,VarBind,NewTerm),
	list_vars(TermList,VarBind,NewTermList).


% generate_vb(+VarList,+VarBind,-NewVarBind)
%
% forall Element from VarList a new Prolog variable is generated and
% the list [Element, NewVariable] is concatenated to VarBind to form NewVarBind 

:- pred generate_vb(list(top),list(top),list(top)).

generate_vb([],VarBind,VarBind).
generate_vb([V|VarList],VarBind,[[V,_NewVar]|VarBind2]) :-
	generate_vb(VarList,VarBind,VarBind2).
	
% find_vb(+Element,+VarBind,-Variable)
%
% searches the entry [Element,Variable] in list VarBind for a given Element 

:- pred find_vb(top,list(top),top).

find_vb(A,_,A) :-
	var(A),!.
find_vb(A,[[A,V1]|_],V1).
find_vb(A,[_|VB],V) :- 
	find_vb(A,VB,V).

:- pred transform_func(dfg_clause,list(top),list(top),list(top),dfg_clause).

transform_func(Clause,Sorts,Preds,Funcs,NewClause) :-
	Clause =.. [Functor|Args],
	member(Functor,Preds),!,
	exchange(Args,Sorts,Funcs,NewArgs,NewFunc),!,
	( NewFunc == [] ->
		NewClause =.. [Functor|NewArgs];
		(NewPred =.. [Functor|NewArgs],
		 append(NewFunc,[NewPred],AndArgs),
		 NewClause =.. [and|AndArgs])).
	
transform_func(Clause,Sorts,Preds,Funcs,NewClause) :-
	Clause =.. [Functor|Args],!,
	transform_funclist(Args,Sorts,Preds,Funcs,NewArgs),
	NewClause =.. [Functor|NewArgs].

transform_func(Clause,_,_,_,Clause).

:- pred transform_funclist(list(dfg_clause),list(top),list(top),list(top),dfg_clause).

transform_funclist([],_,_,_,[]).
transform_funclist([Element|List],Sorts,Preds,Funcs,[NewElement|NewList]) :-
	transform_func(Element,Sorts,Preds,Funcs,NewElement),
	transform_funclist(List,Sorts,Preds,Funcs,NewList).

:- pred exchange(list(top),list(top),list(top),list(top),list(top)).

exchange([],_,_,[],[]).	

exchange([Arg|ArgList],Sorts,Funcs,[Arg|NewArgList],FuncList) :-
	Arg =.. [Functor|_Arguments],
	member(Functor,Sorts),!,
	exchange(ArgList,Sorts,Funcs,NewArgList,FuncList).

exchange([Arg|ArgList],Sorts,Funcs,[_NewVar|NewArgList],[fresult(Arg,_NewVar)|FuncList]) :-
	Arg =.. [Functor|_Arguments],
	member(Functor,Funcs),!,
	exchange(ArgList,Sorts,Funcs,NewArgList,FuncList).
	
exchange([Arg|ArgList],Sorts,Funcs,[Arg|NewArgList],FuncList) :-
	exchange(ArgList,Sorts,Funcs,NewArgList,FuncList).

:- pred andor_list(top,top).

andor_list(Clause,or(NewArgs)) :-
	Clause =.. [or | Args],!,
	listproc(Args,NewArgs).

andor_list(Clause,and(NewArgs)) :-
	Clause =.. [and | Args],!,
	listproc(Args,NewArgs).

andor_list(Clause,NewClause) :-
	Clause =.. [Functor|Args],!,
	listproc(Args,NewArgs),
	NewClause =.. [Functor|NewArgs].

andor_list(Clause,Clause) :-
	atom(Clause).

:- pred listproc(list(top),list(top)).

listproc([],[]).

listproc([Element|List],[NewElement|NewList]) :-
	andor_list(Element,NewElement),
	listproc(List,NewList).

:- pred or_assoc(top,top).

or_assoc(or([]),or([])) :- !.
 
or_assoc(or([or(List)|TermList]), or(NewTermList)) :-
        !,
        append(List, TermList, TermList2),
        or_assoc(or(TermList2), or(NewTermList)).
 
or_assoc(or([Term|TermList]), or(NewTermList)) :-
        !, \+ (Term = or(_List)), 
	or_assoc(Term,NewTerm),
        or_assoc(or(TermList), or(TermList2)),
        NewTermList = [NewTerm|TermList2].

or_assoc(Term,NewTerm) :-
	Term =.. [Functor|Args],!,
	or_assoclist(Args,NewArgs),
	NewTerm =.. [Functor|NewArgs].

or_assoc(Term,Term) :-
	atom(Term).

:- pred or_assoclist(list(top),list(top)).

or_assoclist([],[]).
or_assoclist([Element|List],[NewElement|NewList]) :-
	or_assoc(Element,NewElement),
	or_assoclist(List,NewList).

:- pred transform_clause(top,top).

transform_clause(forall(Vars,Term),forall(Vars,NewTerm)) :-
	!, transform_clause(Term,NewTerm).

transform_clause(exists(Vars,Term),exists(Vars,NewTerm)) :-
	!, transform_clause(Term,NewTerm).

transform_clause(Clause,or(Pos)) :-
	Clause =.. [or , Args],
	transform_clause_list(Args,NewArgs),
	split_pos_neg(NewArgs,Pos,Neg),
	Neg == [],!.

transform_clause(Clause,or([not(Term)|Pos])) :-
	Clause =.. [or , Args],
	transform_clause_list(Args,NewArgs),
	split_pos_neg(NewArgs,Pos,Neg),
	length(Neg,1),!,
	Neg = [Term].

transform_clause(Clause,not(and(Neg))) :-
	Clause =.. [or , Args],
	transform_clause_list(Args,NewArgs),
	split_pos_neg(NewArgs,Pos,Neg),
	Pos = [],!.

transform_clause(Clause,or([not(and(Neg))|Pos])) :-
	Clause =.. [or , Args],
	transform_clause_list(Args,NewArgs),
	split_pos_neg(NewArgs,Pos,Neg),!.

transform_clause(Clause,and(NewArgs)) :-
	Clause =.. [and , Args],!,
	transform_clause_list(Args,NewArgs).

transform_clause(not(Args),not(NewArgs)) :-
	!, transform_clause(Args,NewArgs).

transform_clause(implies(Args1,Args2),implies(NewArgs1,NewArgs2)) :-
	!, transform_clause(Args1,NewArgs1),
	transform_clause(Args2,NewArgs2).

transform_clause(implied(Args1,Args2),implied(NewArgs1,NewArgs2)) :-
	!, transform_clause(Args1,NewArgs1),
	transform_clause(Args2,NewArgs2).

transform_clause(equiv(Args1,Args2),equiv(NewArgs1,NewArgs2)) :-
	!, transform_clause(Args1,NewArgs1),
	transform_clause(Args2,NewArgs2).

transform_clause(Clause,Clause).

:- pred transform_clause_list(list(top),list(top)).

transform_clause_list([],[]).
transform_clause_list([Clause|ClauseList],[NewClause|NewClauseList]) :-
	transform_clause(Clause,NewClause),
	transform_clause_list(ClauseList,NewClauseList).

:- pred split_pos_neg(list(top),list(top),list(top)).

split_pos_neg([],[],[]) :- !.

split_pos_neg([not(Term)|ClauseList],Pos,[Term|Neg]) :-
	!, split_pos_neg(ClauseList,Pos,Neg).

split_pos_neg([Term|ClauseList],[Term|Pos],Neg) :-
	!, split_pos_neg(ClauseList,Pos,Neg).

