%----
% Copyright (C) 1996, Gregor Meyer, FernUniversitaet Hagen, Germany.
% This file is part of the system 'Typical'.
% See the file LICENSE for conditions of distribution and use.
%----
%:- multifile (subtype)/1, (suptype)/1, (type)/1, (pred)/1, (type_import)/1, (local)/1.

















%  files.pl
%
%  file handling
%  implements general file-aliasing as in Quintus Prolog

:- ensure_loaded(typeops).

:- type_import(prolog_types).
:- type_import(prolog_meta).
:- type_import(prolog_io).

:- type_import(lib).

:- type filename = atom.   % = top
:- type accessmode = atom. % = top

%------ 
/*
compose_filename(Dir,FName,DirFName).
 
compose_filename('xdir','ab',X), X='xdir/ab'.
compose_filename('xdir/','ab',X), X='xdir/ab'.
compose_filename('','ab',X), X='ab'.
compose_filename('.','ab',X), X='./ab'.
compose_filename(xdir,'a/b',X), X='xdir/a/b'.
*/

:- pred compose_filename(filename,filename,filename).

compose_filename('',FName,FName).
compose_filename(Dir,FName,DirFName) :-
        name(Dir,DirCode),
        name(FName,FNameCode),
        (last(0'/,DirCode) ->
            append(DirCode,FNameCode,DirFNameCode)
        ;   append(DirCode,[0'/|FNameCode],DirFNameCode)
        ),
        (   name(DirFName,DirFNameCode)
	;   append(DirFNameCode,[0'.,0'p,0'l],DirFNameCodePL),
            name(DirFName,DirFNameCodePL)
	).
 


%------ 
% file_search_path(PathAlias, DirectorySpec)
%   see Quintus built-in hook,  manual p. G-48
% 
:- pred file_search_path(atom,filename).
/*  e.g.
library_directory(xdir) corresponds to
    file_search_path(library,xdir)
:- dynamic file_search_path/3.
*/


 
%------ 
% file_dealiased(FA,F) :-
%    FA is a possibly aliased file name,
%    F is the corresponding file name with aliasing resolved.
%    if no aliased file is found: F = FA.
%
:- pred file_dealiased(atom,filename).
%
file_dealiased(FA,F) :-
    compound(FA),
        functor(FA,Alias,1),
        arg(1,FA,FName),
        find_alias_file(Alias,FName,F0),
    !,
    F=F0.
file_dealiased(FA,F) :-
    compound(FA),
    !,  % no aliased files exists
    arg(1,FA,F).
file_dealiased(F,F).


:- pred find_alias_file(atom,filename,filename).

find_alias_file(Alias,File,DirF) :-
    file_search_path(Alias,Dir),
    compose_filename(Dir,File,DirF),     % Dir=a, File=b/c
    file_exists(DirF).

/*  
:- pred find_alias_file(atom,filename,filename,filename).
find_alias_file(Alias,File,Path,F) :-
    find_alias_file(Alias,File,DirF),
    split_fname(DirF,Path,F).            % Path=a/b, F=c
    ^^ not implemented yet
*/




%------ 
% file_exists(+FileName) :-
% 	Succeed if a file exists, otherwise fail
%
:- pred file_exists(filename).
%
%  Quintus manual H.6-4,  library(files)
%		also file_exists(+FileName,+Permission)


%------
% open_aliased(F,Mode,Stream) :-
%   open file named by F with access mode Mode
%   and return Stream. The file name F may be aliased.
%
:- pred open_aliased(filename,accessmode,stream).
/* 
open_aliased(F,Mode,Stream) :- open(F,Mode,Stream). % Quintus
    open/2, open/3  in Quintus already recognize path-aliases 
 
open_aliased(F,Mode,Stream) :-
    file_dealiased(F,F2),
    open(F2,Mode,Stream).
*/

