%----
% Copyright (C) 1996, Gregor Meyer, FernUniversitaet Hagen, Germany.
% This file is part of the system 'Typical'.
% See the file LICENSE for conditions of distribution and use.
%----
%:- multifile (subtype)/1, (suptype)/1, (type)/1, (pred)/1, (type_import)/1, (local)/1.
















%
%  glib.pl
%
%  several library predicates in addition to lib.pl


:- ensure_loaded(typeops).

:- type_import(prolog_types).
:- type_import(prolog_num).
:- type_import(prolog_arith).
:- type_import(prolog_meta).

:- type_import(lib).


%----
% member_open(X,L) :-
%	same as member/2, L may be an open list.
%
:- pred member_open(T,list(@T)).
%
member_open(_,L) :- var(L), !, fail.
member_open(X,[X|_]).
member_open(X,[_|L]) :- member_open(X,L).

%----
% memberchk_open(X,L) :-
%	same as memberchk/2, L may be an open list.
%
:- pred memberchk_open(T,list(@T)).
%
memberchk_open(_,L) :- var(L), !, fail.
memberchk_open(X,[X|_]) :- !.
memberchk_open(X,[_|L]) :- member_open(X,L).

%-------
:- pred var_member(T,list(@T)).

var_member(X,[Y|_]) :- X == Y.
var_member(X,[_|T]) :- var_member(X,T).

%-------
:- pred var_memberchk(T,list(@T)).

var_memberchk(X,[Y|_]) :- X == Y, !.
var_memberchk(X,[_|T]) :- var_memberchk(X,T).


%-------
% member_elim(X,LX,L) :-
%	X is member of list LX
%	L is LX with all occurrences of X removed.
%
:- pred member_elim(T,list(@T),list(T)) .
%
member_elim(X,[X|T],R) :-  R=T.
member_elim(X,[Y|T],[Y|R]) :- member_elim(X,T,R).

%-------
:- pred var_member_elim(T,list(@T),list(T)) .

var_member_elim(X,[X1|T],R) :- X1==X,  R=T.
var_member_elim(X,[Y|T],[Y|R]) :- var_member_elim(X,T,R).

%-------
% insert_uniq( X, L, XL ) :-
%	the list XL contains X and every member of list L,
%	if X is already member of L, then XL = L.

:- pred insert_uniq( T, list(T), list(@T)).

insert_uniq(X,L,XL) :- member(X,L), !, XL = L.
insert_uniq(X,L,[X|L]).  % not member(X,L).


%-------
:- pred insert_uniq_end(T, list(T), list(@T)).
insert_uniq_end(X,L,LX) :- member(X,L), !, LX = L.
insert_uniq_end(X,L,LX) :- append(L,[X],LX).

%-------
% insert_uniq_open( X, L ) :-
%	variant of insert_uniq for open lists (i.e. tail is a variable),
%	if X is new, then it is added to the (new) tail of L.
% 
:- pred insert_uniq_open( T, list(@T)).
%
insert_uniq_open(X,L) :- var(L), !, L = [X|_].
insert_uniq_open(X,[X|_]) :- !.
insert_uniq_open(X,[_|L]) :- insert_uniq_open(X,L).

%-------
% var_remove_duplicates( LD, L ) :-
%       the list L contains every member of list LD without exact '==' dup's
%
:- pred var_remove_duplicates(list(@T),list(@T)).
%
var_remove_duplicates([],[]).
var_remove_duplicates([X|Xs],Ys) :-
        var_member(X,Xs), !,
        var_remove_duplicates(Xs,Ys).
var_remove_duplicates([X|Xs],[X|Ys]) :-
        % not var_member(X,Xs),
        var_remove_duplicates(Xs,Ys).

:- dynamic par_num_counter/1.

:- pred par_num_counter(nat).
%par_num_counter(1001).
par_num_counter(1).

:- pred reset_num_counter.
reset_num_counter :- 
        retract(par_num_counter(_)),
        assert(par_num_counter(1)).

:- pred new_num(nat).
new_num(N) :-
        par_num_counter(N),!,
        N1 is N+1,
        retract(par_num_counter(N)),
        assert(par_num_counter(N1)).
        
%-------
:- pred contains_var(top,top).
contains_var(Any,X) :- var(Any),!,  Any == X.
contains_var(Any,X) :- Any = [Arg|_], contains_var(Arg,X).
contains_var(Any,X) :- Any = [_|Args], !, contains_var(Args,X).
contains_var(Any,X) :- Any =.. [_F,Arg|Args], contains_var([Arg|Args],X).
% type error: contains_var(Any,X) :- Any =.. [_F,Args], contains_var(Args,X).





 
% anti_subst(+After,-Subst,-Before) :-
%       generate a substitution Subst and a term Before
%       such that Subst applied to Before gives After.
%
%       ex.:  anti_subst(p(a,b),[B=b],p(a,B)).
%             anti_subst(p(X,X),[A=X], p(X,A)).
%
:- pred anti_subst(term,list(top),term).
%
anti_subst(Y,Subst,New) :-
        % not restricted to simple(Y)
        Subst = [ New=Y ].
anti_subst(Y,Subst,X) :-
        compound(Y),   
        functor(Y,F,Arity),     % inspect Y
        functor(X,F,Arity),     % make pattern X
        anti_subst_n(Y,Arity,Subst,X),
        Subst=[_|_].   % trivial substitutions unwanted


:- pred anti_subst_n(term,nat,list(top),term).
 
anti_subst_n(_,0,[],_) :- !.
anti_subst_n(Y,N,Subst,X) :-  % write(anti_subst(Y,N,Subst,X)),nl,
        arg(N,Y,Yarg), 
        % either take argument directly or make it substituted
        (   
            SubstN = [],
            Xarg = Yarg
        ;   
            anti_subst(Yarg,SubstN,Xarg)
        ),
        arg(N,X,Xarg),
        N1 is N -1,  
        anti_subst_n(Y,N1,S,X),
        append(SubstN,S,Subst).
 
 
% test
% anti_subst(f(a,b),S,X)   % , substl(S,X,ZZ).
% anti_subst(f(a,f(b)),S,X)  % , substl(S,X,ZZ).




