%----
% Copyright (C) 1996, Gregor Meyer, FernUniversitaet Hagen, Germany.
% This file is part of the system 'Typical'.
% See the file LICENSE for conditions of distribution and use.
%----
%:- multifile (subtype)/1, (suptype)/1, (type)/1, (pred)/1, (type_import)/1, (local)/1.
















% ----------------------------------------------------------------------
% Operators used in type declarations.
:- op(1199, fy, [(sorts), (operators), (quantifiers), (predicates), (functions), (constants )]).
:- op(1098, fy, [(forall),(exists), (subsort), (predicate), (function)]).  % module keywords
%:- op(1097, fy, [(or),(and),(implies),(implied),(equiv),(not),(equal)]).
% ----------------------------------------------------------------------
 
:- ensure_loaded('decl.pl').

:- dynamic pragma_option/1.

tad(X) :-
	member('-v',X),!,
	member_elim('-v',X,Y),
	assert(pragma_option(verbose)),
	tad(Y).

tad(X) :- 
        % add built-in type goal, must be know for valid_type_names
        Prog = prog([goal|_],_,_,_,_),
        get_from(X,Prog), %  Read Prog + Decls from file
	check_decl(Prog), % Check if all Decls are Defined
        process_input(Prog), % Process the Program
        write(bye),nl,
	halt.

process_input(prog(Sorts,Preds,Funcs,Ord,Clauses)) :-
	close_list(Clauses),
	close_list(Ord),
	po_create([xx|Sorts],Ord,PO),
	type_infer_cl(PO,Clauses,Sorts,Preds,Funcs).

get_from([],_Prog) :- !.       
 
get_from([File|FileList],Prog) :- !,
        get_from(File,Prog),   
        get_from(FileList,Prog).
 
get_from(File,Prog) :-
        atom(File), !,
        read_input(File,Prog).  % Read from the source file
 
get_from(File,_Prog) :-
        write('tad>>  invalid file name '),
        write(File), 
        write(' .abort'), nl, abort.
 
:- dynamic io_opened_module/1.
 
% * Load file FileName
read_input(FileName,Prog) :- 
        % don't read a file twice
        absolute_file_name(FileName,AbsName),
        (io_opened_module(AbsName) 
         -> true      % already read
         ;  assert(io_opened_module(AbsName)),
	  (pragma_option(verbose) ->
	   (write(read_input(FileName)), nl);
	      true),
	  read_input2(AbsName,Prog)).
 
read_input2(FileName,Prog) :- 
        open(FileName,read,Stream),
        read_prg(Stream,Prog),
        close(Stream).

read_prg(Stream,Prog) :-
        read(Stream,Clause),
        !,
        % write(got(Clause)),nl,
        ( Clause == end_of_file 
           -> true   % , write(eof)
           ; transform(Clause,Prog), !,
             read_prg(Stream,Prog)
        ).
  
 
% syntactical markers in DFG-syntax

transform(Term,_Program) :-
	member(Term,[end_of_list,list_of_descriptions,list_of_symbols,list_of_declarations,list_of_axioms,list_of_conjectures,end_problem]),!.

transform((begin_problem(_Problem)),_Program) :-
	!.

% import other Files

transform((include(FileName)),Program) :-
	!, read_input(FileName,Program).

% declaration of all sorts/types

transform((sorts Sorts), prog(S,_P,_F,_O,_C)) :- 
	!, store_sorts(Sorts,S).

% declaration of all predicates

transform((predicates Preds), prog(_S,P,_F,_O,_C)) :-
	!,store_predicates(Preds,P).

%declaration of all functions

transform((functions Funcs), prog(_S,_P,F,_O,_C)) :-
	!, store_funcs(Funcs,F).

% operators are currently not handled

transform((operators _Op), _Program) :-
	!.

% quantifiers are currently not handled

transform((quantifiers _Q), _Program) :-
	!.

% predicates with explicit type constraints
% eg. predicate(p, [@T1,@T2], [T1 =< T2]).

transform((Pred),prog(Sorts,Preds,_F,_O,_C)) :-
	Pred =.. [predicate,Pname,SortVarlist,Co],
	length(SortVarlist,Arity),
	valid_pred_name(Pname,Preds),
        \+ test_pred_decl_variant_ngr(Pname,Arity,_,_),  % must be unique
        !,
	trans_pdecl_list(SortVarlist,SortVarlist1,Co1),
	append(Co,Co1,Co2),
	valid_sort_names_list(SortVarlist1,Sorts),
	valid_sort_names_co(Co2,Sorts),
	simplify_pdecl(SortVarlist1,Co2,Co3),
	de_store_pdecl(Pname,Arity,pred,SortVarlist1,Co3).

% predicates with(?) implicit type constraints
% eg. predicate(p, [type,type]).
%      predicate(p, [T1,T2]).

transform((Pred),prog(Sorts,Preds,_F,_O,_C)) :-
	Pred =.. [predicate,Pname,SortVarlist],
	length(SortVarlist,Arity),
	valid_pred_name(Pname,Preds),
        \+ test_pred_decl_variant_ngr(Pname,Arity,_,_),  % must be unique
        !,
	trans_pdecl_list(SortVarlist,SortVarlist1,Co),
	valid_sort_names_list(SortVarlist1,Sorts),
	valid_sort_names_co(Co,Sorts),
	simplify_pdecl(SortVarlist1,Co,Co1),
	de_store_pdecl(Pname,Arity,pred,SortVarlist1,Co1).

% predicates without implicit type constraints
% eg. predicate(p,type,type).

transform((Pred),prog(Sorts,Preds,_F,_O,_C)) :-
	Pred =.. [predicate,Pname|Sortlist],
	length(Sortlist,Arity),
	valid_pred_name(Pname,Preds),
        \+ test_pred_decl_variant_ngr(Pname,Arity,_,_),  % must be unique
        !,
	valid_sort_names_list(Sortlist,Sorts),
	de_store_pdecl(Pname,Arity,pred,Sortlist,[]).

% functions with result type and parameter types (type constructor)
% eg. constructor(cons,[T,list(T)],list(T)).

transform((Constructor),prog(Sorts,_P,Funcs,_O,_C)) :-
	Constructor =.. [constructor,Fname,Params,Result],
	length(Params,Arity),
	valid_func_name(Fname,Funcs),
	insert_end(Fname,Sorts),
	valid_sort_name(Result,Sorts),
	valid_sort_names_list(Params,Sorts),
	de_store_fdecl(Fname,Arity,Params,Result).

% functions with result type and parameter types (type constructor)
% eg.  constructor(0,[],zero)

transform((Constructor),prog(Sorts,_P,_F,_O,_C)) :-
	Constructor =.. [constructor,Fname,[],Result],
	valid_sort_name(Fname,Sorts),
	valid_sort_name(Result,Sorts),
	de_store_fdecl(Fname,0,[],Result).

% functions with result type and parameter types (flexible sort constructor)
% eg. function(cons,list(T),[T,list(T)]).

transform((Func),prog(Sorts,_P,Funcs,_O,_C)) :-
	Func =.. [function,Fname,Params,Result],
	length(Params,Arity),
	valid_func_name(Fname,Funcs),
	valid_sort_name(Result,Sorts),
	valid_sort_names_list(Params,Sorts),
	de_store_fdecl(Fname,Arity,Params,Result).

% flexible sort constructor 
% eg. zero(0).

transform((Sort),prog(Sorts,_P,Funcs,_O,_C)) :-
	Sort =.. [S,Const],
	memberchk(S,Sorts),
	insert_end(Const,Funcs),
	de_store_fdecl(Const,0,[],S).

% subsort relationships A < B  
% eg. subsort(nat,int).

transform(subsort(Sort1, Sort2),prog(_S,_P,_F,O,_C)) :-
	!, insert_end(Sort1 =< Sort2, O).

% formulas
% eg. formula (forall([v_x], equal(V_x,v_x)), name).

transform((Form), prog(_S,_P,_F,_O,Formulas)) :-
	Form =.. [formula,Formula,_Fname],
	!, insert_end(Formula,Formulas).

% or formula (forall([v_x], equal(V_x,v_x))).

transform((Form), prog(_S,_P,_F,_O,Formulas)) :-
	Form =.. [formula,Formula],
	!, insert_end(Formula,Formulas).

% left-overs not in syntax

transform(Term,_Program) :- 
	write('error: cannot process term '),write(Term),nl,abort.

% insert list of sorts in open list and defines operators for each sort 

store_sorts([],_L).
store_sorts([X|Xs],L) :-
	insert_end(X,L),
%	op(1098,fx,X),
	store_sorts(Xs,L).

% insert list of functions in open list and defines operators for each function 

store_funcs([],_L).
store_funcs([X|Xs],L) :-
	insert_end(X,L),
%	op(1098,fx,X),
	store_funcs(Xs,L).

% insert list of predicates in open list

store_predicates([],_L).
store_predicates([X|Xs],L) :-
	insert_end(X,L),
	store_predicates(Xs,L).	

% validate that a predicate name has been declared before

valid_pred_name(Pname,[Pname|_L]) .
valid_pred_name(Pname,[_P|Preds]) :-
	valid_pred_name(Pname,Preds).
valid_pred_name(Pname,_Preds) :- 
	write('error: predicate '),write(Pname),write(' not declared!'),nl,abort.

% validate that a function name has been declared before

valid_func_name(Fname,[Fname|_L]) .
valid_func_name(Fname,[_F|Funcs]) :-
	valid_func_name(Fname,Funcs).
valid_func_name(Fname,_Funcs) :- 
	write('error: function '),write(Fname),write(' not declared!'),nl,abort.

% test if the list of sort names are valid with the declarations

valid_sort_names_list([],_Sorts).
valid_sort_names_list([X|Xs],Sorts) :-
	valid_sort_name(X,Sorts),
	valid_sort_names_list(Xs,Sorts).

% validate that the sort names in list of constraints have been declared

valid_sort_names_co([],_Sorts).
valid_sort_names_co([X=<Y|L],Sorts) :-
	valid_sort_name(X,Sorts),
	valid_sort_name(Y,Sorts),
	valid_sort_names_co(L,Sorts).

% validate that the sort name is a variable, top or has been declared

valid_sort_name(Sort,_Sorts) :-
	var(Sort), !.
valid_sort_name(top,_Sorts) :-
	!.
valid_sort_name(Sort,Sorts) :-
	Sort =.. [F|Args],
	memberchk(F,Sorts),
	!,
	valid_sort_names_list(Args,Sorts).
valid_sort_name(Sort,_Sorts) :-
	write('error: sort '),write(Sort),write(' not declared!'),nl,abort.

	
% insert in open list.
insert_end(X, L) :- var(L), !, L = [X|_NewL].
insert_end(X, [_|L]) :- insert_end(X, L).

% close open list.
close_list([]) :- !.
close_list([_|L]) :- close_list(L).

% test that all declarations get defined afterwards

check_decl(prog(Sorts,Preds,Funcs,_O,_C)) :-
	close_list(Sorts),
	close_list(Funcs),
	close_list(Preds).
/*	check_sort_decls(Sorts),
	check_pred_decls(Preds),
	check_func_decls(Funcs).
*/ 

% confirm that all sorts that are declared have been defined.

check_sort_decls([]) :- !.

check_sort_decls(Sorts) :-
	setof(S,F^(A^(P^(func_decl_variant_ngr0(F,A,P,S)))),DefSorts),
	subset(DefSorts,Sorts),
	subset(Sorts,[goal,top|DefSorts]),!.

check_sort_decls(Sorts) :-
	setof(S,F^(A^(P^(func_decl_variant_ngr0(F,A,P,S)))),DefSorts),
	subtract(Sorts,[goal,top|DefSorts],Remain),
	Remain \== [],
	write('warning: sort(s) '),write(Remain),write(' declared, but not defined!'),nl,!.

check_sort_decls(Sorts) :-
	write('warning: sort(s) '),write(Sorts),write(' declared, but not defined!'),nl.

% confirm that all functions that are declared have been defined.

check_func_decls([]) :- !.

check_func_decls(Funcs) :-
	setof(F,(A^(P^(S^(func_decl_variant_ngr0(F,A,P,S)))),A>0),DefFuncs),
	subset(DefFuncs,Funcs),
	subset(Funcs,DefFuncs),!.

check_func_decls(Funcs) :-
	setof(F,(A^(P^(S^(func_decl_variant_ngr0(F,A,P,S)))),A>0),DefFuncs),
	subtract(Funcs,DefFuncs,Remain),
	Remain \== [],
	write('warning: function(s) '),write(Remain),write(' declared, but not defined!'),nl,!.

check_func_decls(Funcs) :-
	write('warning: function(s) '),write(Funcs),write(' declared, but not defined!'),nl.

% confirm that all predicates that are declared have been defined.

check_pred_decls([]) :- !.

check_pred_decls(Preds) :-
	setof(P,A^(PN^(K^(C^(pred_decl_variant_ngr0(P,A,PN,K,C))))),DefPreds),
	subset(DefPreds,Preds),
	subset(Preds,DefPreds),!.

check_pred_decls(Preds) :-
	setof(P,A^(PN^(K^(C^(pred_decl_variant_ngr0(P,A,PN,K,C))))),DefPreds),
	subtract(Preds,DefPreds,Remain),
	Remain \== [],
	write('warning: predicate(s) '),write(Remain),write(' declared, but not defined!'),nl,!.

check_pred_decls(Preds) :-
	write('warning: predicate(s) '),write(Preds),write(' declared, but not defined!'),nl.


