%----
% Copyright (C) 1996, Gregor Meyer, FernUniversitaet Hagen, Germany.
% This file is part of the system 'Typical'.
% See the file LICENSE for conditions of distribution and use.
%----
%:- multifile (subtype)/1, (suptype)/1, (type)/1, (pred)/1, (type_import)/1, (local)/1.

















:- ensure_loaded(typeops).

:- type_import(prolog_types).
:- type_import(prolog_io).
:- type_import(glib).
:- type_import(csppo).

:- type_import(checking).
:- type_import(titrace).


type pum_id -->
        pum_f ;  % from function declarations
        pum_p.   % from predicate declarations


:- op(700,xfx,=<=).
type pum_constr(T) --> T =<= T.  % really $\preceq^u$

type pum(T) -->   % principal upper matchers over domain T
        pum(pum_id,list(pum_constr(T))).


%----
% simplify_constraints(PO, , ) :- 
%	remove/solve x=<y where deterministic
%	or elements are ground
%       or ...
%	may fail if constraints are not satisfiable.
%
pred simplify_constraints(poset(@T),
        list(po_constraint(@T1)),list(po_constraint(T1))) -> T1 =< T.
%
simplify_constraints(PO,C,Csimp) :-
	member_elim(X=Y,C,C0),
	!,
	true,
	X=Y,
	simplify_constraints(PO,C0,Csimp).
simplify_constraints(PO,C,Csimp) :-
	member_elim(X=<Y,C,C0),
        \+ var(X),   true,
	\+ var(Y),   true, 
	!,
	true,
	po_test_leq(PO,X,Y),
	simplify_constraints(PO,C0,Csimp).
simplify_constraints(PO,C,Csimp) :-
	member_elim(X=<Y,C,C0),
        (   ground(X),
            po_maximal(PO,X)   % \+ po_ordered_up(PO,X)
        ;   ground(Y),
            po_minimal(PO,Y)   % \+ po_ordered_low(PO,Y)
        ),
        !,
	true,
        X=Y,    % must be equal if no other order exists
	simplify_constraints(PO,C0,Csimp).
simplify_constraints(PO,C,Csimp) :-
	member_elim(lub(Y,[X]),C,C0),
        % one-element list
        !,
	true,
        Y = X,
	simplify_constraints(PO,C0,Csimp).
simplify_constraints(_,C,C).  % no further simplifications 



%----
% make_lub_constraints(C,C2) :-
%       input: C, list of  tv =< tv
%       output: C2, list of var =< symbol, var = lub([tv1,tv2,...])
%
pred make_lub_constraints(list(pum_constr(@T)),list(po_constraint(@T))). 
%
make_lub_constraints([],[]) :- !.
make_lub_constraints([X=<=Y|C],[X=<Y|C2]) :-
        ground(Y), !,
        make_lub_constraints(C,C2).
make_lub_constraints([X=<=Y|C],[lub(Y,[X|L2])|C2]) :-
        var(Y), !,
        select_lub_vars(C,Y,L,C1),
        eliminate_var_dup(L,L2),
        make_lub_constraints(C1,C2).
make_lub_constraints(C,_) :- 
	write('internal-error(mlub)':C),nl,
	fail.

pred select_lub_vars(list(pum_constr(@T)),T,list(T),list(pum_constr(T))).

select_lub_vars([],_YY,[],[]).
select_lub_vars([X=<=Y|C],YY,L,C2) :-
        Y == YY,
        !,
        L=[X|L2],
        select_lub_vars(C,YY,L2,C2).
select_lub_vars([X=<=Y|C],YY,L,[X=<=Y|C2]) :-
        select_lub_vars(C,YY,L,C2).


pred eliminate_var_dup(list(@T),list(@T)). 

eliminate_var_dup([],[]).
eliminate_var_dup([X|Xs],X2s) :-
        test_var_member(X,Xs), !,
        eliminate_var_dup(Xs,X2s).
eliminate_var_dup([X|Xs],[X|X2s]) :-
        eliminate_var_dup(Xs,X2s).
 
% test only, no backtracking
pred test_var_member(T,list(@T)).
test_var_member(XX,[X|_]) :- ground(XX), ground(X), X = XX, !.
test_var_member(XX,[X|_]) :- var(XX), var(X), X == XX, !.
test_var_member(XX,[_|L]) :- test_var_member(XX,L).
 
 

% for checks
pred simple_constraints(list(po_constraint(_))).
simple_constraints(_L) :- true, fail.
simple_constraints([]).
simple_constraints([X|Xs]) :-
	simple_co(X),
	simple_constraints(Xs).

pred simple_co(po_constraint(_)).
simple_co(_X) :- true, fail.
simple_co(X=<Y) :- simple(X),simple(Y).
simple_co(lub(X,L)) :- simple(X),simplist(L).

pred simplist(list(_)).
simplist(_L) :- true, fail.
simplist([]).
simplist([X|Xs]) :- simple(X), simplist(Xs).



 
/*  unused

% e-entailment
% e_entail(C,Co) :- variable constraints C e-entail variable constraints Co
%       C is a list of var(i) =< var(j)
%       C is a list of var(k) =< var(l)
%
pred e_entail(list(po_constr(T)),list(po_constr(@T))).
%
e_entail(C,Co) :-
        poset_from_C(C,L),
        satisfiable(L,Co,_Sub).

*/

/*
vars_in_C(C,V) :- vars_in_C(C,[],V).
vars_in_C([],Acc,AccR) :- !, reverse(Acc,AccR).
vars_in_C([X=<Y|C],Cv0,Cvars) :-
        ( X = var(_I) ->
            insert_uniq(X,Cv0,Cv1)
        ;   Cv1 = Cv0
        ),
        ( Y = var(_J) ->
            insert_uniq(Y,Cv1,Cv2)
        ;   Cv2 = Cv1
        ),
        vars_in_C(C,Cv2,Cvars).
*/



 
/*
% make list (no duplicates) of elems in constraint-list
elems_in_C(C,E) :- elems_in_C(C,[],E).
elems_in_C([],Acc,AccR) :- !, reverse(Acc,AccR).
elems_in_C([X=<Y|C],Cv0,Cvars) :-
        insert_uniq(X,Cv0,Cv1),
        insert_uniq(Y,Cv1,Cv2),
        elems_in_C(C,Cv2,Cvars).

%----
% extend_poset(PO,?  with constraints ?
%
local
pred extend_poset(poset(T),constraints(T),poset(@T)).
% extend_poset(L,C,LC) :- LC is poset L extended with C.
extend_poset(L,C,LC) :-
        elems_in_C(C,TC),
        extend_poset(L,TC,C,LC).
*/


