%----
% Copyright (C) 1996, Gregor Meyer, FernUniversitaet Hagen, Germany.
% This file is part of the system 'Typical'.
% See the file LICENSE for conditions of distribution and use.
%----
%:- multifile (subtype)/1, (suptype)/1, (type)/1, (pred)/1, (type_import)/1, (local)/1.
















:- ensure_loaded(typeops). 

%
% decompose constraints over type terms
%
%

:- type_import(prolog_types).
:- type_import(prolog_num).
:- type_import(prolog_io).
:- type_import(glib).
:- type_import(csppo).
:- type_import(ti_2po).
:- type_import(ttlib).
:- type_import(ttord).
:- type_import(cllib).



type subst_flag --> subst.   % if some substitution is applied




%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% decompose constraints on type terms


% decompose_simpum_co(PO,BV,PUMs,Co, BV9,PUMs9,Co9) :-
%	PUMs and Co are exhaustively decomposed
%
:- pred decompose_simpum_co(poset(@T),
	list(T), list(ttx_pum),ttx_co,
	list(T), list(pum(T)),list(po_constraint(T))).
%
decompose_simpum_co(PO,BV,PUMs,Co, BV9,PUMs9,Co9) :-
        true,
	length(PUMs,N),
	Bound is N*1+2,
	decompose_simpum_co_b(Bound,PO,BV,PUMs,Co, BV9,PUMs9,Co9).
	


:- pred decompose_simpum_co_b(int,poset(@T),
	list(T), list(ttx_pum),ttx_co,
	list(T), list(pum(T)),list(po_constraint(T))).
%
decompose_simpum_co_b(Bound,_PO,_BV,_PUMs,_Co, _BV9,_PUMs9,_Co9) :-
	Bound < 0,
	!,
	write('decompose sim-constraints':forced_termination),nl,
	fail.
decompose_simpum_co_b(Bound,PO,BV,PUMs,Co, BV9,PUMs9,Co9) :-
	decompose_simpum(PO,BV,PUMs, _Sub,BV2,PUMs2   ),
        true,
	true,
	decompose_constraints(BV2,Co,  _Subst,Co2  ),
   	true,
	true,
	(   % _Subst == subst ->
	    % a substitution during dec_co possibly sets
	    % PUMs2 into a non-reduced form 
	    (\+ simple_pums(PUMs2)) ->
	    true,
	    true,
	    Bound2 is Bound - 1,
      	    decompose_simpum_co_b(Bound2,PO,BV2,PUMs2,Co2, BV9,PUMs9,Co9)
	;
 	    BV9 = BV2, PUMs9 = PUMs2, Co9=Co2
	).




% decompose_contraints(BV, List of exploded type-term-constraints, CN).
%	type-term = log.var or [Func|Args]
%	result CN: normalized constraints
%
:- pred decompose_constraints(list(_),list(top),subst_flag,list(po_constraint(_))).
decompose_constraints(BV,C,Subst,CN) :- 
        dec_co_exhaustive(BV,C,Subst,CN).

%---
% decompose constraints over type terms into simple constraints
%
:- pred dec_co_exhaustive(list(_),ttx_co,subst_flag,list(po_constraint(_))).
%
dec_co_exhaustive(BV,S,Subst,S9) :-
	length(S,N),
	Bound is N * 5,     % arbitrary bound to ensure termination
	dec_co_exhaustive_bounded(Bound,BV,S,Subst,S9),
	no_complex_co_terms(S9).
	

:- pred dec_co_exhaustive_bounded(int,list(_),ttx_co,subst_flag,list(po_constraint(_))).

dec_co_exhaustive_bounded(Bound,_BV,_S,_,_S9) :-
	Bound < 0,
	!,
	write('decompose constraints':forced_termination),nl,
	fail.
dec_co_exhaustive_bounded(Bound,BV,S,Subst,S9) :-
        true, true,
        dec_co1(BV,S,Subst,S2), !,
	Bound1 is Bound - 1,
        dec_co_exhaustive_bounded(Bound1,BV,S2,Subst,S9).
dec_co_exhaustive_bounded(_N,_BV,S,_Subst,S) :-  % not dec_co1
        true.

%---
% dec_co1: make one decomposition step

:- pred dec_co1(list(_),ttx_co,subst_flag,ttx_co).

dec_co1(_BV, [_X=<Y|C], _Subst,C2) :-
        \+ var(Y),
	Y=top,
	!,
	C2=C.

dec_co1(_BV, [X=<Y|C], _Subst,C2) :-
        \+ var(X), \+ var(Y),
        X = [FX|ArgsX], Y = [FY|ArgsY],
        true,  true,
        !, true,
        make_f_pairs(=<,ArgsX,ArgsY,CA),
        append([FX=<FY|CA],C,C2).

dec_co1(_BV, [lub(X,[Y,Z])|C], Subst,C2) :-    % compound = y lub z
        \+ var(X),  % Y and Z may be var or compound terms
        X = [FX|ArgsX],      true,
	((var(Y);var(Z)) -> Subst = subst),
	Y = [FY|ArgsY], Z = [FZ|ArgsZ],
	gen_list(ArgsX,ArgsY),        % fail if ArgsY has diff. length
	gen_list(ArgsX,ArgsZ),        % ...
        true,  
        true,
        !, true,
	make_list_pairs(ArgsY,ArgsZ,ArgsYZ),
        make_f_pairs(lub,ArgsX,ArgsYZ,CA),
        append([lub(FX,[FY,FZ])|CA],C,C2).

 
dec_co1(_BV, [lub(X,[Y,Z])|C], Subst, C2) :-        % x = compound lub z
        \+ var(Y), 
        Y = [FY|ArgsY],      true,
	((var(X);var(Z)) -> Subst = subst),
        X = [FX|ArgsX], Z = [FZ|ArgsZ],
        gen_list(ArgsY,ArgsX),        % fail if ArgsX has diff. length 
        gen_list(ArgsY,ArgsZ),        % ...
        true,  
        true,
        !, true,
        make_list_pairs(ArgsY,ArgsZ,ArgsYZ),
        make_f_pairs(lub,ArgsX,ArgsYZ,CA),
        append([lub(FX,[FY,FZ])|CA],C,C2).

  
dec_co1(_BV, [lub(X,[Y,Z])|C], Subst, C2) :-        % x = y lub compound
        \+ var(Z), 
        Z = [FZ|ArgsZ],      true, 
	((var(X);var(Y)) -> Subst = subst),
        X = [FX|ArgsX], Y = [FY|ArgsY], 
        gen_list(ArgsZ,ArgsX),        % fail if ArgsX has diff. length  
        gen_list(ArgsZ,ArgsY),        % ... 
        true,   
        true, 
        !, true, 
        make_list_pairs(ArgsY,ArgsZ,ArgsYZ), 
        make_f_pairs(lub,ArgsX,ArgsYZ,CA), 
        append([lub(FX,[FY,FZ])|CA],C,C2). 
 
 
dec_co1(_BV, [X=Y|C], _Subst,C2) :-
        \+ var(X), \+ var(Y),
        X = [FX|ArgsX], Y = [FY|ArgsY],
        true,  true,
        !, true,
        make_f_pairs(=,ArgsX,ArgsY,CA),
        append([FX=FY|CA],C,C2).
 
dec_co1(BV, [X=<_Y|C], _Subst,C2) :-     % narrow  'bot =< term'
        var(X),
	var_member(X,BV),
        !,  true,
	true,
        C2 = C.

dec_co1(_BV, [X=<Y|C], subst,C2) :-     % narrow  'var =< term'
        var(X), \+ var(Y), 
        Y = [FY|ArgsY],   true,  % FY not nec. ground
        !,  true,
        gen_list(ArgsY,VX),
        true,  % show old X-var
        X = [FX|VX],
        make_f_pairs(=<,VX,ArgsY,CA),
        append([FX=<FY|CA],C,C2).

 
dec_co1(_BV, [X=Y|C], subst,C2) :-     % narrow  'var = term'
        var(X), \+ var(Y),
        Y = [FY|ArgsY],   true,  % FY not nec. ground
        !,  true,
        gen_list(ArgsY,VX),
        true,  % show old X-var
        X = [FX|VX],
        make_f_pairs(=,VX,ArgsY,CA),
        append([FX=FY|CA],C,C2).
 

dec_co1(_BV, [X=<Y|C], subst,C2) :-       % narrow 'term =< var'
        \+ var(X), var(Y),
        X = [FX|ArgsX],   true,  % FX not nec. ground
        !, true,
        gen_list(ArgsX,VY),
        true,
        Y = [FY|VY],
        make_f_pairs(=<,ArgsX,VY,CA),
        append([FX=<FY|CA],C,C2).

 
dec_co1(_BV, [X=Y|C], subst,C2) :-       % narrow 'term = var'
        \+ var(X), var(Y),
        X = [FX|ArgsX],   true,  % FX not nec. ground
        !, true,
        gen_list(ArgsX,VY),
        true,
        Y = [FY|VY],
        make_f_pairs(=,ArgsX,VY,CA),
        append([FX=FY|CA],C,C2).


dec_co1(BV, [Q|C], Subst,[Q|C2]) :-
        dec_co1(BV,C,Subst,C2).




%-------------------------------------------------

% decompose principal upper matcher into reduced form


% decompose_simpum(PO,BV, M, BV99,C99) :- M is a multiset of pum-constraints
%	decompose simultaneous p.u.m.
%
:- pred decompose_simpum(poset(@X),list(X),list(ttx_pum), subst_flag, list(X), list(pum(X))).
%
decompose_simpum(PO, BV, P, Subst,BV0,P9) :-
	length(P,N),
	Bound is N *2,
	decompose_simpum_b(Bound,PO, BV, P, Subst,BV0,P9).


:- pred decompose_simpum_b(int,poset(@X),list(X),list(ttx_pum), subst_flag, list(X), list(pum(X))).
decompose_simpum_b(Bound,_PO,_BV,_P, _Subst,_BV9,_P9) :-
        Bound < 0,
        !,
        write('decompose simpum':forced_termination),nl,
        fail.
decompose_simpum_b(_N,_, BV, [], _Subst,BV, []).
decompose_simpum_b(Bound,PO,BV,[P|M], _Subst,BV9, P9) :-  true,
	decompose_pum(PO,BV,P, BV2,P2),
	P2=pum(_,L),
	true,
	decompose_simpum_b(Bound,PO,BV2,M, _Sub,BV3,P3),
	append([P2],P3,P4),
	% type-error append(P2,P3,P4),
	(   no_complex_terms(L) ->
	    BV9 = BV3, P9=P4
	;
	    true,
	    Bound2 is Bound - 1,
	    decompose_simpum_b(Bound2,PO,BV3,P4,_XSubst,BV9,P9)
	).


:- local
pred decompose_pum(poset(@X),list(X),ttx_pum, list(X),ttx_pum).
%
decompose_pum(_PO,BV, pum(Pid,S), BV9, pum(Pid,S9)) :- 
	true,
	true, 
	!,
	dec_pum_exhaustive(BV, pum(Pid,S), _Subst, BV9, pum(_P,S9)),
	true,
	% S9 should not contain complex type terms
	% otherwise pum does not exist
	no_complex_terms(S9).

:- local
pred no_complex_terms(list(pum_constr(_))).
no_complex_terms([]).
no_complex_terms([X=<=Y|C]) :-
	(var(X) ; atom(X)),
	(var(Y) ; atom(Y)),
	no_complex_terms(C).

:- local
pred no_complex_co_terms(list(po_constraint(_))).
no_complex_co_terms([]).
no_complex_co_terms([X=<Y|C]) :-
        (var(X) ; atom(X)),
        (var(Y) ; atom(Y)),
        no_complex_co_terms(C).
	
:- local
pred dec_pum_exhaustive(list(X),ttx_pum,subst_flag,list(@X),ttx_pum).
%
dec_pum_exhaustive( BV,PS, Subst, BV9,PS9) :-
	PS = pum(_,S),
        length(S,N),
        Bound is N * 3,     % 'arbitrary' bound to ensure termination
        dec_pum_exhaustive_bounded(Bound,BV,PS, Subst,BV9,PS9).

:- local 
pred dec_pum_exhaustive_bounded(int,list(@X),ttx_pum,subst_flag,list(X),ttx_pum).
%
dec_pum_exhaustive_bounded(Bound,_BV,_S, _Subst,_BV9,_S9) :-
        Bound < 0,
        !,
        write('decompose upper-matcher':forced_termination),nl,
        fail.
dec_pum_exhaustive_bounded(Bound,BV,S, Subst,BV9,S9) :-
	true, true,
	dec_pum1(BV,S, Subst,BV2,S2), !,
	Bound1 is Bound -1,
	dec_pum_exhaustive_bounded(Bound1,BV2,S2, Subst,BV9,S9).
dec_pum_exhaustive_bounded(_N,BV,S, _Subst,BV2,S) :-  % no more dec_pum1
	true,
	(   S = pum(pum_f,S0) ->
	    select_botvars(BV,S0,BV2),
	    true
	;   true,
	    BV2=BV
	).

% select_botvars(BV,PUM,BV9) :-
% 	BV9 is the list of (bottom-)vars from BV
%	that do not occur in some rhs in PUM
:- local
pred select_botvars(list(@X),list(top),list(X)).
select_botvars([],_PUM,[]).
select_botvars([B|Bs],PUM,BL) :-
	var_in_pum_rhs(B,PUM),
	!,
	select_botvars(Bs,PUM,BL).
select_botvars([B|Bs],PUM,[B|BL]) :-
	% \+ var_in_pum_rhs(B,PUM),
	% !,
	select_botvars(Bs,PUM,BL).
	

% var_in_pum_rhs(X,PUM) :-
%	X is a (bottom) variable
%	and does occur in right hand side of some =<= in PUM
%
:- local
pred var_in_pum_rhs(top,list(top)).	
%
var_in_pum_rhs(Z,[_X=<=Y|_C]) :-
	Z == Y.
var_in_pum_rhs(Z,[_|C]) :-
	var_in_pum_rhs(Z,C).


% dec_pum1(BV,pum(PF,C), Subst,BV',pum(PF,C')) :-
%	BV are bottom-vars,
%	a p.u.m. transformation step applies to C,
%	BV' is new set of bottom-vars,
%	C' is the transformed C,
%       if some substitution was used then Subst = subst
%	else Subst is unchanged (a free var)
%

:- local
pred dec_pum1(list(@X),ttx_pum, subst_flag, list(X),ttx_pum).
dec_pum1(BV, pum(PF,[_X=<=Y|C]), _Subst, BV, pum(PF,C2)) :-     % ... =<= top
	\+ var(Y),
	Y=top,
	!,
	C2=C.

% compund =<= compound
dec_pum1(BV, pum(PF,[X=<=Y|C]), _Subst, BV, pum(PF,C2)) :-     
	\+ var(X), \+ var(Y),
	X = [FX|ArgsX], Y = [FY|ArgsY],
	true, true,
	!, true, 
	make_f_pairs(=<=,ArgsX,ArgsY,CA),
	append([FX=<=FY|CA],C,C2).

% bottom =<= ...
dec_pum1(BV, pum(PF,[X=<=Y|C]), _Subst, BVmore, pum(PF,C2)) :-     
        var(X),
        var_memberchk(X,BV),
        !,  true,
	true,
	(   PF = pum_f ->
	    variables_of(Y,VY),
	    append(VY,BV,BVmore)
	;   BVmore = BV
	),
	true,
        C2 = C.

% narrow  var =<= compound
dec_pum1(BV, pum(PF,[X=<=Y|C]), subst, BV, pum(PF,C2)) :-    
        var(X), \+ var(Y), 
        Y = [FY|ArgsY],   
        !,  true,
            true,  true,
        gen_list(ArgsY,VX),
        true,
        X = [FX|VX],
        make_f_pairs(=<=,VX,ArgsY,CA),
        append([FX=<=FY|CA],C,C2).
 
% narrow  compound =<= var
dec_pum1(BV, pum(PF,[X=<=Y|C]), subst, BV, pum(PF,C2)) :-   
	\+ var(X), var(Y), 
	X = [FX|ArgsX],
	!, true,   true,
	gen_list(ArgsX,VY),
	true,
	Y = [FY|VY], 
	make_f_pairs(=<=,ArgsX,VY,CA),
	append([FX=<=FY|CA],C,C2).

dec_pum1(BV, pum(PF,[Q|C]), Subst, BV2, pum(PF,[Q|C2])) :-
	dec_pum1(BV,pum(PF,C), Subst, BV2,pum(PF,C2)).





