%----
% Copyright (C) 1996, Gregor Meyer, FernUniversitaet Hagen, Germany.
% This file is part of the system 'Typical'.
% See the file LICENSE for conditions of distribution and use.
%----
%:- multifile (subtype)/1, (suptype)/1, (type)/1, (pred)/1, (type_import)/1, (local)/1.
















:- ensure_loaded(typeops).

:- type_import(prolog_types).
:- type_import(glib).
:- type_import(csppo).
:- type_import(decl).
:- type_import(ti_2po).
:- type_import(poset).

% type terms with partial order


%----------------------------------------------

:- pred make_lequm_pairs(list(T),list(T),list(pum_constr(@T))).
%
make_lequm_pairs([],[],[]).
make_lequm_pairs([X|Xs],[Y|Ys],[X=<=Y|XYs]) :-
        make_lequm_pairs(Xs,Ys,XYs).

 
:- predl make_lequm_bottom_pairs(list(T),list(pum_constr(@T))).
%
make_lequm_bottom_pairs([],[]).
make_lequm_bottom_pairs([Y|Ys],[bottom=<=Y|XYs]) :-
        make_lequm_bottom_pairs(Ys,XYs).
 

:- pred make_lequm_multipairs(list(T),list(T),list(list(pum_constr(@T)))).
%
make_lequm_multipairs([],[],[]).
make_lequm_multipairs([X|Xs],[Y|Ys],[[X=<=Y]|XYs]) :-
        make_lequm_multipairs(Xs,Ys,XYs).
 

%pred make_f_pairs(top,list(T),list(T),list(top)).  no good
% types of second and third argument may be totally unrelated, 
% e.g. int and list(int)
:- pred make_f_pairs(top,list(_),list(_),list(top)).
%
make_f_pairs(_Func,[],[],[]).
make_f_pairs(Func,[X|Xs],[Y|Ys],[FXY|FXYs]) :-
        FXY =.. [Func,X,Y],
        make_f_pairs(Func,Xs,Ys,FXYs).


:- pred make_f_pairs_const(top,list(_T),top,list(top)).
%
make_f_pairs_const(_Func,[],_Y,[]).
make_f_pairs_const(Func,[X|Xs],Y,[FXY|FXYs]) :-
        FXY =.. [Func,X,Y],
        make_f_pairs_const(Func,Xs,Y,FXYs).


%
% make_list_pairs([a1,a2,a3,...],[b1,b2,b3,...],LL) :-
% 	LL = [ [a1,b1], [a2,b2], [a3,b3] ... ]
%
:- pred make_list_pairs(list(T),list(T),list(list(@T))).
%
make_list_pairs([],[],[]).
make_list_pairs([X|Xs],[Y|Ys],[[X,Y]|FXYs]) :-
        make_list_pairs(Xs,Ys,FXYs).


:- predl make_f_bottom_pairs(top,list(_T),list(top)).
%
make_f_bottom_pairs(_Func,[],[]).
make_f_bottom_pairs(Func,[Y|Ys],[FbY|FbYs]) :-
        FbY =.. [Func,bottom,Y],
        make_f_bottom_pairs(Func,Ys,FbYs).


:- local
type unused_par --> par.

% leqtt on type terms
%
:- pred leqtt(poset(_),top,top).
%
%leqtt(_PO, par(I),par(I)).
leqtt(PO, X,Y) :- X =..[Xfunc|Xargs], true,
	      Y =..[Yfunc|Yargs], true,
              po_test_leq(PO, Xfunc,Yfunc),
              leqtt_args(PO, Xargs,Yargs).

:- local
pred leqtt_args(poset(_),list(top),list(top)).
%
leqtt_args(_PO, [],[]).
leqtt_args(PO, [X|Xs],[Y|Ys]) :- leqtt(PO, X,Y), leqtt_args(PO, Xs,Ys).



% glbtt on type terms
%
:- predl glbtt(poset(_),top,top,top).
%
%glbtt(PO, X, Y, Z) :- write(call(glbtt(po,X,Y,Z))), fail.
glbtt(_PO, top,Y,Z) :- !, Z = Y.
glbtt(_PO, X,top,Z) :- !, Z = X.
glbtt(PO, bottom,_,B) :- !, po_indomain(PO,bottom), B = bottom.
glbtt(PO, _,bottom,B) :- !, po_indomain(PO,bottom), B = bottom.
%glbtt(_PO, par(I),par(I),par(I)).
glbtt(PO, X,Y,Z) :- % validate X,Y neither top nor bottom
		X =..[Xfunc|Xargs], true,
                Y =..[Yfunc|Yargs], true,
                po_glb(PO, Xfunc,Yfunc,Zfunc),
                glbtt_args(PO, Xargs,Yargs,Zargs),
                Z =..[Zfunc|Zargs].

:- local
pred glbtt_args(poset(_),list(top),list(top),list(top)).
%
glbtt_args(_PO, [],[],[]).
glbtt_args(PO, [X|Xs],[Y|Ys],[Z|Zs]) :- 
	glbtt(PO, X,Y,Z),
	glbtt_args(PO, Xs,Ys,Zs).


/* unused

% lubtt on type terms
%
:- predl lubtt(poset(_),top,top,top).
%
lubtt(PO, bottom,X,X) :- po_indomain(PO,bottom).
lubtt(PO, X,bottom,X) :- po_indomain(PO,bottom).
%lubtt(_PO, par(I),par(I),par(I)).
lubtt(PO, X,Y,Z) :- X =..[Xfunc|Xargs], check_top(\+ Xfunc = par),
                Y =..[Yfunc|Yargs], check_top(\+ Yfunc = par),
                po_lub(PO, Xfunc,Yfunc,Zfunc),
                lubtt_args(PO, Xargs,Yargs,Zargs),
                Z =..[Zfunc|Zargs].
 
:- local
pred lubtt_args(poset(_),list(top),list(top),list(top)).
%
lubtt_args(_PO, [],[],[]).
lubtt_args(PO, [X|Xs],[Y|Ys],[Z|Zs]) :- 
	lubtt(PO, X,Y,Z),
	lubtt_args(PO, Xs,Ys,Zs).
*/


:- pred nonempty(poset(top), top).
%
nonempty(PO,_T) :- \+ po_indomain(PO,bottom), !, true.
nonempty(PO,T) :- nonempty_rec(PO,T).

:- local
pred nonempty_rec(poset(_),top).
%
nonempty_rec(_,T) :- (T = []; T = [_|_]), true.
nonempty_rec(_PO,T) :-
	T =.. [_F],  % a base type
	\+ T=bottom, !.
nonempty_rec(PO,T) :-
	% test if there is a functor with type T.
	func_decl_variant_ngr(_F,ArgsD,T),  
	nonempty_rec_list(PO,ArgsD), !.
nonempty_rec(PO,T) :-
	T =.. [F|Args],
	po_less(PO,Fsub,F),  true,
	Tsub =..[Fsub|Args],
	nonempty_rec(PO,Tsub).

:- local
pred nonempty_rec_list(poset(_),list(top)).
%
nonempty_rec_list(_PO,[]).
nonempty_rec_list(PO,[T|Ts]) :- 
	nonempty_rec(PO,T),
	nonempty_rec_list(PO,Ts).










